use {
    super::*,
    std::{
        fmt::{self, Write},
        io::{self, Read},
    },
};

#[derive(Debug)]
pub enum EncoderError {
    IO(io::Error),
    FMT(fmt::Error),
}

impl From<io::Error> for EncoderError {
    fn from(value: io::Error) -> Self {
        Self::IO(value)
    }
}

impl From<fmt::Error> for EncoderError {
    fn from(value: fmt::Error) -> Self {
        Self::FMT(value)
    }
}

pub struct Encoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B32Alphabet,
    wrap: Option<usize>,
}

impl<R: Read, W: Write> Encoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B32Alphabet>, wrap: Option<usize>) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
            wrap,
        }
    }

    pub fn encode(&mut self) -> Result<(), EncoderError> {
        let mut total: usize = 0;
        loop {
            let mut buf = [0; 5];
            let mut num: u64 = 0;
            let n_bytes;
            loop {
                n_bytes = match self.reader.read(&mut buf) {
                    Ok(n) => n,
                    Err(e) if e.kind() == io::ErrorKind::Interrupted => continue,
                    Err(e) => return Err(e.into()),
                };
                break;
            }
            if n_bytes == 0 {
                break;
            }
            for (idx, c) in buf.iter().enumerate() {
                num <<= 8;
                if idx < n_bytes {
                    num |= *c as u64;
                }
            }
            let mut buf = ['\0'; 8];
            for i in (0..8).rev() {
                let b = (num & 0x1f) as u8;
                buf[i] = char::try_from(self.alphabet.items[b as usize]).unwrap();
                num >>= 5;
            }
            let mut outlen = n_bytes * 8 / 5;
            let rem = n_bytes * 8 % 5;
            if rem != 0 {
                outlen += 1;
            }
            buf[0..outlen].iter().try_for_each(|c| {
                self.writer.write_char(*c)?;
                total += 1;
                if let Some(wrap) = self.wrap {
                    if total % wrap == 0 {
                        self.writer.write_char('\n')?;
                    }
                }
                Ok::<(), fmt::Error>(())
            })?;
            if let Some(pad) = self.alphabet.pad {
                if rem != 0 {
                    for _c in 0..8 - outlen {
                        self.writer.write_char(char::try_from(pad).unwrap())?;
                    }
                }
            }
            if n_bytes < 5 {
                break;
            }
        }
        Ok(())
    }
}

impl<R: Read> Encoder<R, String> {
    pub fn output(self) -> String {
        self.writer
    }
}

#[cfg(test)]
mod test {
    use {
        super::*,
        std::{
            fs::{self, File},
            io::BufReader,
        },
    };

    #[test]
    fn hello() {
        let reader = "Hello, World".as_bytes();
        let writer = String::new();
        let mut encoder = Encoder::new(reader, writer, None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.output(), "JBSWY3DPFQQFO33SNRSA====");
        encoder = Encoder::new("Hello, World!".as_bytes(), String::new(), None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.output(), "JBSWY3DPFQQFO33SNRSCC===");
        encoder = Encoder::new("Hello, World!\n".as_bytes(), String::new(), None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.output(), "JBSWY3DPFQQFO33SNRSCCCQ=");
    }

    #[test]
    fn encode_from_file() {
        let infile = File::open("src/testdata/lorem.txt").unwrap();
        let reader = BufReader::new(infile);
        let outfile = fs::read_to_string("src/testdata/lorem_b32.txt").unwrap();
        let mut encoder = Encoder::new(reader, String::new(), None, Some(76));
        encoder.encode().unwrap();
        assert_eq!(encoder.output(), outfile);
    }
}
